/* Copyright (c) 2009, Cornell University.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer
       in the documentation and/or other materials provided with the
       distribution.
     * Neither the name of the University nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "config.h"
#include <ally.h>
#include <stdlib.h>

typedef struct ally_agree_fab_paxos_t ally_agree_fab_paxos_t;

struct ally_agree_fab_paxos_t
{
  ally_agree_t super;
  ally_peers_t *proposers;
  ally_peers_t *acceptors;
  ally_peers_t *learners;
  ally_agree_fab_paxos_role_t role;
  ally_agree_learn_t learn;
  size_t leader;
  size_t self;
};

static int
message (ally_agree_fab_paxos_t *fab, size_t size, const char *message)
{
  /* FIXME: perform agreement protocol, learn only in learners at end */
  fab->learn ((ally_agree_t *) fab, size, message);
  /*
    leader
      start
	"7:propose,4:name,5:round,5:value,8:progress," -> all acceptors
    proposer
      start
	wait for timeout
	if not satisfied
	  suspect the leader
      learned from learner i
	learned[i] = true
	if |{j where learned[j]}| >= ceil((l + f + 1) / 2)
	  "9:satisfied," -> all proposers
      satisfied from proposer i
	satisfied[i] = true
    acceptor
      propose(name,round,value,progress) from leader
	if not already accepted then
	  accepted = name,round,value
	"8:accepted,4:name,5:round,5:value," -> all learners
    learner
      start
	wait for timeout
	while not learned
	  "4:pull," -> all learners
      pull from learner i
	if learned
	  "8:learned,4:name,5:round,5:value," -> learner i
      learned(name,round,value) from learner i
	learn[i] = name,round,value
	if |{j where learn[j] = name,round}| >= f + 1
	  learned = name,round,value
      accepted(name,round,value) from acceptor i
	accepted[i] = name,round,value
	if |{j where accepted[j] = name,round}| >= ceil((a + 3f + 1) / 2)
	  learned = name,round,value
	  "7:learned," -> all proposers
  */
  return -1;
}

static int
propose (ally_agree_fab_paxos_t *fab, size_t size, const char *message)
{
  size_t i;
  int err;
  /* FIXME: send to proposers, not learners directly */
  for (i = 0; i < fab->learners->size; ++i)
    {
      err = fab->learners->send (fab->learners, i, size, message);
      if (err)
	return err;
    }
  return 0;
}

static void
done (ally_agree_fab_paxos_t *fab)
{
  free (fab);
}

ally_agree_t *
ally_agree_fab_paxos (ally_peers_t *proposers, ally_peers_t *acceptors,
                      ally_peers_t *learners,
                      ally_agree_fab_paxos_role_t role,
                      ally_agree_learn_t learn, size_t self)
{
  ally_agree_fab_paxos_t *fab = malloc (sizeof *fab);
  if (fab)
    {
      fab->super.message = (ally_agree_message_t) message;
      fab->super.propose = (ally_agree_propose_t) propose;
      fab->super.done = (ally_agree_done_t) done;
      fab->proposers = proposers;
      fab->acceptors = acceptors;
      fab->learners = learners;
      fab->learn = learn;
      fab->role = role;
      fab->self = self;
      fab->leader = 0;          /* FIXME: elect a leader */
    }
  return (ally_agree_t *) fab;
}
